"""

"""

import json
import logging
from dataclasses import dataclass
from datetime import datetime

import aiohttp
from bs4 import BeautifulSoup as bs4
from feedparser import FeedParserDict, parse

log = logging.getLogger(__name__)
dumps = lambda _dict: json.dumps(_dict, indent=8)


@dataclass
class Article:
    """Represents a news article, or entry from an RSS feed."""

    title: str | None
    description: str | None
    url: str | None
    published: datetime | None
    author: str | None

    @classmethod
    def from_entry(cls, entry:FeedParserDict):
        """Create an Article from an RSS feed entry.

        Parameters
        ----------
        entry : FeedParserDict
            An entry pulled from a complete FeedParserDict object.

        Returns
        -------
        Article
            The Article created from the feed entry.
        """

        log.debug("Creating Article from entry: %s", dumps(entry))

        published_parsed = entry.get("published_parsed")
        published = datetime(*entry.published_parsed[0:-2]) if published_parsed else None

        return cls(
            title=entry.get("title"),
            description=entry.get("description"),
            url=entry.get("link"),
            published=published,
            author = entry.get("author")
        )

    async def get_thumbnail_url(self) -> str | None:
        """Returns the thumbnail URL for an article.

        Returns
        -------
        str or None
            The thumbnail URL, or None if not found.
        """

        log.debug("Fetching thumbnail for article: %s", self)

        async with aiohttp.ClientSession() as session:
            async with session.get(self.url) as response:
                html = await response.text()

        soup = bs4(html, "html.parser")
        image_element = soup.select_one("meta[property='og:image']")
        return image_element.get("content") if image_element else None


@dataclass
class Source:
    """Represents an RSS source."""
    
    name: str | None
    url: str | None
    icon_url: str | None
    feed: FeedParserDict

    @classmethod
    def from_parsed(cls, feed:FeedParserDict):
        """Returns a Source object from a parsed feed.

        Parameters
        ----------
        feed : FeedParserDict
            The feed used to create the Source.

        Returns
        -------
        Source
            The Source object
        """

        log.debug("Creating Source from feed: %s", dumps(feed))

        return cls(
            name=feed.get("channel", {}).get("title"),
            url=feed.get("channel", {}).get("link"),
            icon_url=feed.get("feed", {}).get("image", {}).get("href"),
            feed=feed
        )

    def get_latest_articles(self, max: int) -> list[Article]:
        """Returns a list of Article objects.

        Parameters
        ----------
        max : int
            The maximum number of articles to return.

        Returns
        -------
        list of Article
            A list of Article objects.
        """

        log.debug("Fetching latest articles from %s, max=%s", self, max)

        return [
            Article.from_entry(entry)
            for i, entry in enumerate(self.feed.entries)
            if i < max
        ]


def get_source(rss_url: str) -> Source:
    """_summary_

    Parameters
    ----------
    rss_url : str
        _description_

    Returns
    -------
    Source
        _description_
    """

    parsed_feed = parse(rss_url)
    return Source.from_parsed(parsed_feed)
